//-------------------------------------------------------------------------------------------------+
// GlobalPathPlanner.cpp
// oHvRTC̃CNX̊֐`
//-------------------------------------------------------------------------------------------------+

#include "GlobalPathPlanner.h"
#include <cmath>
#include <iostream>
#include <mmsystem.h>

#define MINIMUM_VALUE	0.001	// oH̍œKɗp鏬߂̒l[m]

using namespace std;


//-------------------------------------------------------------------------------------------------+
// Module specification
//-------------------------------------------------------------------------------------------------+
static const char* globalpathplanner_spec[] =
  {
    "implementation_id", "GlobalPathPlanner",
    "type_name",         "GlobalPathPlanner",
    "description",       "Planning Global Path",
    "version",           "1.0.0",
    "vendor",            "aisl.cs.tut.ac.jp",
    "category",          "GlobalPlannin",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.RobotSafetyRadius", "1.0",
    "conf.default.RobotVelocity", "0.5",
//  "conf.default.ConstA", "1.356",
    "conf.default.ConstB", "0.341",
    // Widget
    "conf.__widget__.RobotSafetyRadius", "text",
    "conf.__widget__.RobotVelocity", "text",
//  "conf.__widget__.ConstA", "text",
    "conf.__widget__.ConstB", "text",
    // Constraints
    ""
  };


//-------------------------------------------------------------------------------------------------+
// RXgN^
//-------------------------------------------------------------------------------------------------+
GlobalPathPlanner::GlobalPathPlanner(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_startIn("start", m_start),
    m_goalIn("goal", m_goal),
    m_pathOut("path", m_path),
    m_messageOut("message", m_message),
    m_AreaInfoServicePort("AreaInfoService")
{

	cerr << "------------------------------------------------------------+\n"
		 << " Global Path Planner Component (Ver. 11th February 2012)\n"
		 << "------------------------------------------------------------+\n\n";
}


//-------------------------------------------------------------------------------------------------+
// fXgN^
//-------------------------------------------------------------------------------------------------+
GlobalPathPlanner::~GlobalPathPlanner(){
}


//-------------------------------------------------------------------------------------------------+
// vON̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t GlobalPathPlanner::onInitialize(){

	// Set InPort buffers
	addInPort( "start", m_startIn );
	addInPort( "goal", m_goalIn );

	// Set OutPort buffer
	addOutPort( "path", m_pathOut );
	addOutPort("message", m_messageOut);

	// Set service consumers to Ports
	m_AreaInfoServicePort.registerConsumer( "AreaInfoService", "TUT::AreaInfoService", m_AreaInfoService );
  
	// Set CORBA Service Ports
	addPort( m_AreaInfoServicePort );

	// Bind variables and configuration variable
	bindParameter( "RobotSafetyRadius", m_RobotSafetyRadius, "1.0" );
	bindParameter( "RobotVelocity", m_RobotVelocity, "0.5" );
//	bindParameter( "ConstA", m_ConstA, "1.356" );
	bindParameter( "ConstB", m_ConstB, "0.341" );

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// ̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t GlobalPathPlanner::onActivated(RTC::UniqueId ec_id){

	no_start = no_goal = true;

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// 񊈐̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t GlobalPathPlanner::onDeactivated(RTC::UniqueId ec_id){

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// s郁C[`
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t GlobalPathPlanner::onExecute(RTC::UniqueId ec_id){

	if( !CheckConnectServicePort() ) return RTC::RTC_OK;				// T[rXڑ

	if( !m_startIn.isNew() && !m_goalIn.isNew() ) return RTC::RTC_OK;	// Vf[^Ȃ

	if( m_startIn.isNew() ){

		while( !m_startIn.isEmpty() ) m_startIn.read();
		no_start = false;
	}

	if( m_goalIn.isNew() ){

		while( !m_goalIn.isEmpty() ) m_goalIn.read();
		no_goal = false;
	}

	if( no_start || no_goal ) return RTC::RTC_OK;

	// ~
	if( m_goal.data.length() == 0 ){

		m_path.data.length(0);
		
		m_pathOut.write();

		cerr << "Number of Generated waypoints = 0\n";

		m_message.data = "GlobalPathPlanner:SUCCESS";
		m_messageOut.write();

		cerr << "An empty path was output to stop the robot.\n";

		return RTC::RTC_OK;
	}

	m_message.tm.sec  = timeGetTime() / 1000;
	m_message.tm.nsec = ( timeGetTime() % 1000 ) * 1000000;

	Coordinate start( m_start.data.position.x, m_start.data.position.y );
	Coordinate goal( m_goal.data[0].position.x, m_goal.data[0].position.y );

	cerr << "Compute Global Path from ( "<< start.x <<", " << start.y << " ) to ( " << goal.x << ", " << goal.y << " ).\n";

	SubAreas subareas = ConvertSubAreas();

	Network network( subareas, m_RobotVelocity, m_ConstB );

	if( network.SubAreaID( start ) == VOID_SUBAREA_ID || network.SubAreaID( goal ) == VOID_SUBAREA_ID ){

		m_path.data.length( 0 );
		m_pathOut.write();

		cerr << "Generated no waypoints (Coordinate is out of area)\n";

		m_message.data = "GlobalPathPlanner:FAILED";
		m_messageOut.write();

		return RTC::RTC_OK;
	}

	PathCost path = network.PathFinding( start, goal );

	if( path.path.empty() && network.SubAreaID( start ) != network.SubAreaID( goal ) ){

		m_path.data.length( 0 );
		m_pathOut.write();

		cerr << "Generated no waypoints (no path)\n";

		m_message.data = "GlobalPathPlanner:FAILED";
		m_messageOut.write();

		return RTC::RTC_OK;
	}

	Lines borders = ConvertPath( path.path );

	vector<Coordinate> points = Optimisation( ConvertPath( borders, start, goal ) );

	m_path = ConvertPath( points );

	m_pathOut.write();

	cerr << "Number of Generated waypoints = " << points.size() -2 << endl;

	cerr << "route:\n";
	for( int i=0; i<(int)points.size(); i++ ){
		cerr << "( "<< points[i].x <<", " << points[i].y << " )\n";
	}
	cerr << endl;

	m_message.data = "GlobalPathPlanner:SUCCESS";
	m_messageOut.write();

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// oH̃f[^`ϊ
//-------------------------------------------------------------------------------------------------+
Lines GlobalPathPlanner::ConvertPath( Path path ){

	Lines global_path;

	while( !path.empty() ){

		SubArea current = path.front();
		path.pop_front();

		if( !path.empty() ) global_path.push_back( current.Border( path.front() ) );
	}

	return global_path;
}


//-------------------------------------------------------------------------------------------------+
// oH̃f[^`ϊ
//-------------------------------------------------------------------------------------------------+
vector<Coordinate> GlobalPathPlanner::ConvertPath( Lines path, Coordinate start, Coordinate goal ){

	vector<Coordinate> global_path;

	global_path.push_back( start );

	// ܂͋E̒ԓ_i[
	for( int i=0; i<(int)path.size(); i++ ) global_path.push_back( path[i].Center() );

	global_path.push_back( goal );


	Coordinate waypoint = start;

	for( int i=1; i<(int)global_path.size()-1; i++ ){

		// ʘH͐^񒆂ʂ΂悢
		if( path[i-1].Length() <= 2.0 * m_RobotSafetyRadius ) continue;

		// P^񒆂ŋߓ_ɕύX
		if( i < (int)path.size() ){
			global_path[i+1] = path[i].PointToLine( global_path[i-1], m_RobotSafetyRadius ).Inverse();
		}

		Line connect( global_path[i-1], global_path[i+1] );

		Coordinate cross = connect.Kreuzung( path[i-1] );

		global_path[i] = path[i-1].PointToLine( cross, m_RobotSafetyRadius ).Inverse();
	}

	return global_path;
}


//-------------------------------------------------------------------------------------------------+
// oH̃f[^`ϊ
//-------------------------------------------------------------------------------------------------+
vector<Coordinate> GlobalPathPlanner::ConvertPath( Path path, Coordinate start, Coordinate goal ){

	vector<Coordinate> global_path;

	global_path.push_back( start );

	Coordinate waypoint = start;

	while( !path.empty() ){

		SubArea current = path.front();
		path.pop_front();

		if( path.empty() ) break;

		Line border = current.Border( path.front() );

		if( border.Length() <= 2.0 * m_RobotSafetyRadius ) waypoint = border.Center();
		else waypoint = border.PointToLine( waypoint, m_RobotSafetyRadius ).Inverse();

		global_path.push_back( waypoint );
	}

	global_path.push_back( goal );

	return global_path;
}


//-------------------------------------------------------------------------------------------------+
// ꒼ɂȂ񂾒ԖړInČoHœK
//-------------------------------------------------------------------------------------------------+
vector<Coordinate> GlobalPathPlanner::Optimisation( vector<Coordinate> path ){

	vector<Coordinate> optimised;

	optimised.push_back( path.front() );

	for( int i=1; i<(int)path.size()-1; i++ ){

		Line line( path[i-1], path[i+1] ); 

		if( MINIMUM_VALUE < line.PointToLine( path[i] ).Length() ) optimised.push_back( path[i] );
	}

	optimised.push_back( path.back() );

	return optimised;
}


//-------------------------------------------------------------------------------------------------+
// oH̃f[^`ϊ
//-------------------------------------------------------------------------------------------------+
TimedPoseVel2DSeq GlobalPathPlanner::ConvertPath( vector<Coordinate> path ){

	TimedPoseVel2DSeq global_path;

	global_path.data.length( path.size() -1 );

	path.push_back( path[path.size()-1] );

	for( int i=1; i<(int)path.size()-1; i++){

		RTC::PoseVel2D waypoint;

		waypoint.pose.position.x = path[i].x;
		waypoint.pose.position.y = path[i].y;
		waypoint.pose.heading    = atan2( path[i+1].y - path[i].y, path[i+1].x - path[i].x );
		waypoint.velocities.vx   = m_RobotVelocity * cos( waypoint.pose.heading );
		waypoint.velocities.vy   = m_RobotVelocity * sin( waypoint.pose.heading );
		waypoint.velocities.va   = 0.0;

		global_path.data[i-1] = waypoint;
	}

	// ŏIn_iS[ĵ߂̐ݒ
	global_path.data[global_path.data.length()-1].pose.heading  = m_goal.data[0].heading;
	global_path.data[global_path.data.length()-1].velocities.va = 0.0;
	global_path.data[global_path.data.length()-1].velocities.vx = 0.0;
	global_path.data[global_path.data.length()-1].velocities.vy = 0.0;

	return global_path;
}


//-------------------------------------------------------------------------------------------------+
// T[rX|[g瓾AreaInfoTuGAɕϊ
//-------------------------------------------------------------------------------------------------+
SubAreas GlobalPathPlanner::ConvertSubAreas(){

	SubAreas subareas;

	if( !CheckConnectServicePort() ) return subareas;

	TUT::TimedAreaInfo *areainfo = m_AreaInfoService->getAreaInfo();

	for( int i=0; i< areainfo->data.length(); i++ ){

		Vertexes vert;

		for( int v=0; v<areainfo->data[i].vertexes.length(); v++ ){

			vert.push_back( Coordinate( areainfo->data[i].vertexes[v].x, areainfo->data[i].vertexes[v].y ) );
		}

		SubArea area( areainfo->data[i].id, vert );

		area.population = areainfo->data[i].population;

		for( int l=0; l<areainfo->data[i].links.length(); l++ ){

			Coordinate c[2];

			c[0].x = areainfo->data[i].links[l].boder.p1.x;
			c[0].y = areainfo->data[i].links[l].boder.p1.y;
			c[1].x = areainfo->data[i].links[l].boder.p2.x;
			c[1].y = areainfo->data[i].links[l].boder.p2.y;

			area.AddLink( areainfo->data[i].links[l].id, Line( c[0], c[1] ) );
		}

		subareas.push_back( area );
	}

	delete areainfo;

	return subareas;
}


//-------------------------------------------------------------------------------------------------+
// T[rX|[g̐ڑmF
// ڑĂtrueԂ
//-------------------------------------------------------------------------------------------------+
bool GlobalPathPlanner::CheckConnectServicePort(){

	if( CORBA::is_nil(m_AreaInfoService.getObject()) ){

		cerr << "Service port is not conected with service provider.\n";
		return false;
	}

	if( m_AreaInfoService.getObject()->_non_existent() ){
	
		cerr << "Service provider is not activated.\n";
		return false;
	}

	return true;
}



//-------------------------------------------------------------------------------------------------+
extern "C"
{
 
  void GlobalPathPlannerInit(RTC::Manager* manager)
  {
    coil::Properties profile(globalpathplanner_spec);
    manager->registerFactory(profile,
                             RTC::Create<GlobalPathPlanner>,
                             RTC::Delete<GlobalPathPlanner>);
  }
  
};


