//-------------------------------------------------------------------------------------------------+
// GlobalPathPlanner.h
// oHvRTC̃CNX`
//-------------------------------------------------------------------------------------------------+

#ifndef GLOBALPATHPLANNER_H
#define GLOBALPATHPLANNER_H

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>
#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// Service Consumer stub headers
#include "AislGlobalPlanningStub.h"
#include "IIS2Stub.h"

#include "Simulator/Network.h"

using namespace RTC;
using namespace IIS;


//-------------------------------------------------------------------------------------------------+
// GlobalPathPlanner
//-------------------------------------------------------------------------------------------------+
class GlobalPathPlanner : public RTC::DataFlowComponentBase{
public:

	GlobalPathPlanner(RTC::Manager* manager);
	~GlobalPathPlanner();
	virtual RTC::ReturnCode_t onInitialize();
	virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);
	virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);
	virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

protected:
	double m_RobotSafetyRadius;
	double m_RobotVelocity;
//	double m_ConstA;
	double m_ConstB;
	IIS::TimedPose2D m_start;
	InPort<IIS::TimedPose2D> m_startIn;
	IIS::TimedPose2DSeq m_goal;
	InPort<IIS::TimedPose2DSeq> m_goalIn;

	// DataOutPort declaration
	TimedPoseVel2DSeq m_path;
	OutPort<TimedPoseVel2DSeq> m_pathOut;

	TimedString m_message;
	OutPort<TimedString> m_messageOut;

	// CORBA Port declaration
	RTC::CorbaPort m_AreaInfoServicePort;

	// Consumer declaration
	RTC::CorbaConsumer<TUT::AreaInfoService> m_AreaInfoService;

private:
	bool no_start, no_goal;

	Lines ConvertPath( Path path );
	vector<Coordinate> ConvertPath( Lines path, Coordinate start, Coordinate goal );
	vector<Coordinate> ConvertPath( Path path, Coordinate start, Coordinate goal );
	vector<Coordinate> Optimisation( vector<Coordinate> path );
	TimedPoseVel2DSeq ConvertPath( vector<Coordinate> path );
	SubAreas ConvertSubAreas();
	bool CheckConnectServicePort();
};



//-------------------------------------------------------------------------------------------------+
extern "C"
{
  DLL_EXPORT void GlobalPathPlannerInit(RTC::Manager* manager);
};

#endif // GLOBALPATHPLANNER_H
